/**
  *  \file client/dialogs/simoptions.cc
  *  \brief Simulator Options Dialog
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/simoptions.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "game/sim.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "util/strtab.h"
#include "client/widgets/optiongrid.h"
#include "ui/widgets/quit.h"

namespace {
    class WSimOptionTile : public WOptionGrid {
        GSimOptions& opts;

        static const char*const left_column[];
        static const char*const right_column[];
        static const char keys[];

        void chooseMode();
     public:
        inline WSimOptionTile(GSimOptions& opts);

        void init();
        void onClick(int id);
        void render();
    };
}

/* Strings are enumerated here to simplify layout computation */
const char*const WSimOptionTile::left_column[] = {
    N_("Mode"),
    N_("Engine-Shield Bonus"),
    N_("Fed Crew Bonus"),
    N_("Random Left/Right"),
    N_("Left/Right Balance"),
    N_("Honor alliances/teams"),
    N_("Battles per simulation"),
    N_("Seed control"),
    N_("Randomize FCodes")
};

const char WSimOptionTile::keys[] = { 'M', 'E', 'C', 'R', 'L', 'A', 'B', 'S', 'F' };

/* Note: for simplicity, we assume that the first entries
   match VcrMode */
const char*const WSimOptionTile::right_column[] = {
    /* Host mode */
    N_("Host"),                 // 0
    N_("PHost 2"),
    N_("PHost 3"),
    N_("PHost 4"),
    N_("FLAK"),
    N_("NuHost"),               // 5
    0,
    0,
    0,
    0,

    /* ES-Bonus */
    N_("none"),                 // 10
    /* Boolean options */
    N_("Yes"),                  // 11
    N_("No"),                   // 12
    /* Balancing */
    N_("360 kt (Host)"),        // 13
    N_("Master at Arms"),       // 14
    /* battles */
    N_("one"),                  // 15
    N_("complete"),             // 16
    /* Random fcode */
    N_("once"),                 // 17
    N_("every time")            // 18
};

WSimOptionTile::WSimOptionTile(GSimOptions& opts)
    : WOptionGrid(getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_column, countof(left_column))),
                  getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_column, countof(right_column)))),
      opts(opts)
{
    init();
}

void
WSimOptionTile::init()
{
    CTASSERT(countof(left_column) == countof(keys));
    for (size_t i = 0; i < countof(left_column); ++i) {
        addItem(keys[i], keys[i], _(left_column[i]));
    }
    find('M').setFont(FONT_NORMAL_BOLD);
    sig_click.add(this, &WSimOptionTile::onClick);
    render();
}

void
WSimOptionTile::onClick(int id)
{
    switch (id) {
     case 'E':
        // ESB
        {
            UIDecimalSelector select(0, 0, 500, 10);
            select.setValue(opts.engine_shield_bonus);
            if (select.doStandardDialog(_("Engine-Shield Bonus"), _("Rate [%]"), "pcc2:simopts")) {
                opts.engine_shield_bonus = select.getValue();
                render();
            }
        }
        break;
     case 'C':
        // Fed bonus
        opts.scotty_bonus = !opts.scotty_bonus;
        render();
        break;
     case 'R':
        // random l/r
        opts.random_left_right = !opts.random_left_right;
        render();
        break;
     case 'L':
        // balancing
        if (int(opts.balancing_mode) >= opts.BalanceMAX) {
            opts.balancing_mode = GSimOptions::BalancingMode(0);
        } else {
            opts.balancing_mode = GSimOptions::BalancingMode(opts.balancing_mode + 1);
        }
        render();
        break;
     case 'P':
     case 'M':
        // host
        chooseMode();
        render();
        break;
     case 'A':
        // allies
        opts.honor_alliances = !opts.honor_alliances;
        render();
        break;
     case 'B':
        // battles per sim
        opts.only_one_simulation = !opts.only_one_simulation;
        if (!opts.only_one_simulation)
            opts.seed_control = false;
        render();
        break;
     case 'S':
        // seed control
        opts.seed_control = !opts.seed_control;
        if (opts.seed_control) {
            opts.only_one_simulation = true;
        }
        render();
        break;
     case 'F':
        // random fcodes
        opts.randomize_fcodes_on_every_fight = !opts.randomize_fcodes_on_every_fight;
        render();
        break;
    }
}

void
WSimOptionTile::render()
{
    // Mode
    find('M').setValue(_(right_column[opts.vcr_mode]));

    // ESB
    if (opts.engine_shield_bonus != 0) {
        find('E').setValue(format("%d%%", opts.engine_shield_bonus));
    } else {
        find('E').setValue(_(right_column[10]));
    }

    // Crew Bonus
    find('C').setValue(_(right_column[opts.scotty_bonus ? 11 : 12]));

    // Random left/right
    find('R').setValue(_(right_column[opts.random_left_right ? 11 : 12]));

    // Left/right balance
    find('L').setValue(opts.balancing_mode == GSimOptions::Balance360k
                       ? _(right_column[13])
                       : opts.balancing_mode == GSimOptions::BalanceMasterAtArms
                       ? _(right_column[14])
                       : _(right_column[10]));
    // Honor alliances
    find('A').setValue(_(right_column[opts.honor_alliances ? 11 : 12]));

    // Battles per simulation
    find('B').setValue(_(right_column[opts.only_one_simulation ? 15 : 16]));

    // Seed control
    find('S').setValue(_(right_column[opts.seed_control ? 11 : 12]));

    // Random fcodes
    find('F').setValue(_(right_column[opts.randomize_fcodes_on_every_fight ? 18 : 17]));
}

void
WSimOptionTile::chooseMode()
{
    // Build list of modes
    StringList sl;
    for (int i = 0; i <= GSimOptions::VcrMAX; ++i) {
        // do we accept this mode?
#ifndef CONF_FLAK_SUPPORT
        if (i == GSimOptions::VcrFLAK)
            continue;
#endif
        sl.add(i, _(WSimOptionTile::right_column[i]));
    }

    // Ask user
    UIStandardListbox box(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    box.scrollToKey(opts.vcr_mode);
    if (box.doStandardListbox(_("Mode"), "pcc2:simopts")) {
        opts.setMode(GSimOptions::VcrMode(box.getKey()));
    }
}

/** Edit simulator options.
    \param opts [in/out] Options
    \retval true Options were possibly changed, dialog confirmed
    \retval false Options not changed, dialog cancelled */
bool
editSimulatorOptions(GSimOptions& opts)
{
    // Temporary copy to work on
    GSimOptions work_opts = opts;

    // Build dialog
    ObjectHolder h;

    UIWindow& win(h.add(new UIWindow(_("Simulator Options"), UIWindow::BLUE_WINDOW, COLOR_GRAY)));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(h.add(new WSimOptionTile(work_opts)));

    UIGroup& group_buttons(h.add(new UIGroup(UIHBoxLayout::instance5)));
    group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));
    group_buttons.add(h.add(new UIButton(_("Help"), 'h')));
    win.add(group_buttons);
    win.add(h.add(new WHelpWidget("pcc2:simopts")));
    win.add(h.add(new UIQuit(cm_Escape)));
    win.pack();
    ui_root->centerWidget(win);

    // Run it
    int rv = ui_root->runModal(win);
    if (rv == cm_Enter) {
        opts = work_opts;
        return true;
    } else {
        return false;
    }
}
