/**
  *  \file client/scr-msg.cc
  *  \brief Message-related User Interface Stuff
  *
  *  This file implements the send/revise/read message screens.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/wedit.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/team.h"
#include "game/playerset.h"
#include "client/dialogs/outboxdlg.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/setselect.h"
#include "client/dialogs/msgreceiver.h"
#include "ui/widgets/quit.h"

namespace {
    class WEditMessageWindow : public UIWindow {
        ObjectHolder h;
        WEditor      editor;
        GOutbox&     outbox;
        GPlayerSet   recv;
     public:
        WEditMessageWindow(string_t text, GPlayerSet receivers, GOutbox& outbox);

        void setText(string_t text);
        void setReceivers(GPlayerSet receivers);

        bool handleEvent(const UIEvent& e, bool second_pass);

        string_t getText() const
            { return editor.getContentAsString(editor.getUserMinimumLine(), editor.getUserMaximumLine()); }
        GPlayerSet getReceivers() const
            { return recv; }
    };
}

WEditMessageWindow::WEditMessageWindow(string_t text, GPlayerSet receivers, GOutbox& outbox)
    : UIWindow(_("Edit Message"), UIWindow::BLUE_WINDOW, COLOR_BLACK),
      editor(0),
      outbox(outbox),
      recv(receivers)
{
    setLayout(UIVBoxLayout::instance5);

    editor.insertString(outbox.getHeadersForDisplay(receivers));
    editor.setMaxLineLength(40);
    editor.setMaxStoredLines(24);
    /* a message may have 18 lines. Two are taken by the headers. */
    editor.setUserLineRestrictions(editor.getCurrentLine(), 20);
    if (!text.empty())
        setText(text);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UIButton(1, _("F10 - Send"), SDLK_F10)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"),     SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), ss_Alt + 'h')));
    add(editor);
    add(g);
    add(h.add(new WHelpWidget("pcc2:msgout")));
    add(h.add(new UIQuit(0)));
    pack();
}

void
WEditMessageWindow::setText(string_t text)
{
    for (int i = editor.getUserMinimumLine(); i < 24 && !text.empty(); ++i) {
        editor.setLine(i, strFirst(text, "\n"));
        if (!strRemove(text, "\n"))
            break;
    }
}

void
WEditMessageWindow::setReceivers(GPlayerSet receivers)
{
    if (receivers == this->recv)
        return;

    WEditor::line_t old_limit = editor.getUserMinimumLine(), index;

    this->recv = receivers;
    string_t new_header = outbox.getHeadersForDisplay(receivers);

    for (index = 0; !new_header.empty(); ++index) {
        if (index >= old_limit) {
            editor.insertLine(index);
            ++old_limit;
        }
        editor.setLine(index, strFirst(new_header, "\n"));
        strRemove(new_header, "\n");
    }

    if (index < old_limit)
        editor.deleteLine(index, old_limit - index);

    editor.setUserLineRestrictions(index, 20);
    drawWidget();
}

bool
WEditMessageWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == ss_Alt + 't') {
            WPlayerSetSelector sel(0, NUM_OWNERS, 15*font_ems[FONT_NORMAL]);
            sel.setSelectedPlayers(getReceivers());
            if (chooseMessageReceiver(sel, _("Change Receiver"), "pcc2:msgout")) {
                setReceivers(sel.getSelectedPlayers());
            }
            return true;
        } else if (e.key.code == ss_Alt + 'f') {
            // FIXME: change 'From' if playing multiple races
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
doSendOrReviseMessage()
{
    GOutbox& outbox = getDisplayedTurn().getOutbox(::getPlayerId());

    WPlayerSetSelector sel(0, NUM_OWNERS, 15*font_ems[FONT_NORMAL]);
    switch (chooseNewMessageReceiver(sel, outbox.getCount())) {
     case nmsgConfirm:
        doSendMessage(string_t(), sel.getSelectedPlayers());
        break;
     case nmsgRevise:
        doOutboxScreen();
        break;
     case nmsgCancel:
        break;
    }
}

void
doSendMessage(string_t text, GPlayerSet receivers)
{
    GOutbox& outbox = getDisplayedTurn().getOutbox(::getPlayerId());

    WEditMessageWindow win(text, receivers, outbox);
    ui_root->centerWidget(win);
    ui_root->add(win);
    bool rv = ui_root->run();
    ui_root->close(win);
    if (rv) {
        outbox.addMessage(win.getText(), win.getReceivers());
        outbox.sig_sel_changed.raise();
        messageBox(format(_("%d message%!1{s%} sent this turn."), outbox.getCount()),
                   _("Message sent"));
    }
}

void
doEditMessage(GOutbox& outbox, int index)
{
    WEditMessageWindow win(outbox.getRawText(index), outbox.getReceiverMask(index), outbox);
    ui_root->centerWidget(win);
    if (ui_root->runModal(win)) {
        outbox.setMessage(index, win.getText(), win.getReceivers());
        outbox.sig_sel_changed.raise();
    }
}
