/**
  *  \file game/pvcr.h
  *  \brief PVCR Implementation
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PVCR_H
#define PCC_V2_GAME_PVCR_H

#include "game/vcr.h"

#define PVCR_INTEGER

#define PVCR_PREPARED_RNG

class VcrPlayerPHost : public VcrPlayer {
 public:
    enum {
        VCR_MAX_BEAMS = 20,
        VCR_MAX_TORPS = 20,
        VCR_MAX_FTRS  = 50,
        VCR_MAX_BAYS  = 50      // 50 allowed on planets, 20 on ship
    };
 private:
    struct RegularFormula;
    struct AlternativeFormula;
    typedef double damage_t;
    /** Changing values. */
    struct RunningStatus {
        uint16_t beam_status[VCR_MAX_BEAMS];        ///< Beam status, [0, ~1000]
        uint16_t torp_status[VCR_MAX_TORPS];        ///< Torpedo launcher status, [0, ~1000]
        uint16_t bay_status[VCR_MAX_BAYS];          ///< Fighter bay status, [0, ~1000]
        uint8_t  fighter_status[VCR_MAX_FTRS];      ///< Fighter status, [0, 2]
        uint16_t fighter_shots_left[VCR_MAX_FTRS];  ///< Fighter strikes remaining
        int  fighter_x[VCR_MAX_FTRS];           ///< Fighter X position, in meters
        int  obj_x;                             ///< X position of baseship, in meters
        int  launched_fighters;                 ///< Number of fighters currently out
        int  launch_countdown;                  ///< Countdown to next fighter launch
        GVcrObject obj;
#ifndef PVCR_INTEGER
        damage_t shield;        ///< Shield status, [0, 100]
        damage_t damage;        ///< Damage, [0, 150]
        damage_t crew;          ///< Crew, [0, 100]
#else
        int32 shield_scaled;    ///< Shield status, scaled with scale
        int32 damage_scaled2;   ///< Damage, scaled with scale*100
        int32 crew_scaled2;     ///< Crew, scaled with scale*100
#endif
    };

#ifdef PVCR_PREPARED_RNG
    struct PreparedRNG {
        uint32_t divi;
        uint32_t limit;

        void operator=(uint32_t max);
    };
#endif

    /** Precomputed values which do not change. */
    struct FixedStatus {
#ifdef PVCR_PREPARED_RNG
        PreparedRNG beam_recharge;
        PreparedRNG bay_recharge;
        PreparedRNG torp_recharge;
#else
        int beam_recharge;      ///< Precomputed gross beam recharge rate.
        int torp_recharge;      ///< Precomputed torp recharge rate
        int bay_recharge;       ///< Precomputed bay recharge rate
#endif
        int beam_hit_odds;      ///< Precomputed gross beam hit odds (%)
        int torp_hit_odds;      ///< Precomputed gross torp hit odds (%)
        int side;
#ifdef PVCR_INTEGER
        int32 scale;            ///< Scale factor for status value
        int32 max_scaled;       ///< Maximum damage value. 10000 * scale
        int32 damage_limit_scaled; ///< Scaled damage limit, scaled with 100*scale
        int32_t mass_plus1;
#else
        double mass_plus1;
        int damage_limit;       ///< Damage limit (150 or 100).
#endif
        /// \name Cached and adjusted values of combat configuration options.
        //@{
        int ShieldDamageScaling;
        int ShieldKillScaling;  // new 4.0
        int HullDamageScaling;
        int CrewKillScaling;
        int MaxFightersLaunched;
        int StrikesPerFighter;
        int BayLaunchInterval;
        int FighterMovementSpeed;
        int FighterBeamExplosive;
        int FighterBeamKill;
        int FighterFiringRange;
        int BeamHitFighterRange;
        int BeamHitFighterCharge;
        int BeamFiringRange;
        int BeamHitShipCharge;
        int TorpFiringRange;
        int ShipMovementSpeed;
        ///@}
    };

    struct Status {
        FixedStatus f;
        RunningStatus r;
    };

    struct StatusToken : public VcrStatus {
        Status status[2];
        int32_t seed;
        int status_word;
        bool done;
        int one_f, right_probab;
        StatusToken(time_type t);
    };

    struct DetStatus {
#ifdef PVCR_INTEGER
        int32 shield_scaled, damage_scaled, crew_scaled;
#else
        double shield, damage, crew;
#endif
        int fighters, torps;
    };

    Status status[2];
    uint32_t seed;
    bool done;
    int one_f, right_probab;
    unsigned capabilities;
    DetStatus det_status[2];
    bool det_valid;
    time_type det_timer;

    /** Ruleset selection. True if PHost 3 or later. */
    bool phost3;

    inline uint32 random64k();
#ifdef PVCR_PREPARED_RNG
    inline int randomRange(const PreparedRNG& rng);
#else
    inline int randomRange(uint32_t n);
#endif
    inline int randomRange100();
    inline bool randomRange100LT(int comp);

    template<typename Formula>
    static inline bool hitT(Status& st, int kill, int expl, bool is_death_ray);
    bool hit(Status& st, int kill, int expl, bool is_death_ray);

    static inline int computeBayRechargeRate(int num, const GVcrObject& obj);
    void fighterRecharge(Status& st);
    void fighterLaunch(Status& st);
    void fighterMove(Status& st);
    void fighterIntercept();
    bool fighterAttack(Status& st, Status& opp);

    static inline int computeBeamHitOdds(int beam, const GVcrObject& obj);
    static inline int computeBeamRechargeRate(int beam, const GVcrObject& obj);
    void beamRecharge(Status& st);
    int beamFindNearestFighter(const Status& st, const Status& oppst);
    bool beamFire(Status& st, Status& opp);

    static inline int computeTorpHitOdds(int torp, const GVcrObject& obj);
    static inline int computeTubeRechargeRate(int torp, const GVcrObject& obj);
    void torpsRecharge(Status& st);
    bool torpsFire(Status& st, Status& opp);

    void moveObjects();
    bool canStillFight(const Status& st, const Status& opp);

    /// \name Inactivity Detection
    //@{
    void initActivityDetector();
    bool compareDetStatus(const DetStatus& a, const Status& st);
    void fillInDetStatus(DetStatus& a, const Status& st);
    bool checkCombatActivity();
    //@}

    bool checkVcrSide(GClassicVcr& vcr, int side);
 public:
    VcrPlayerPHost(bool phost3);
    bool checkVcr(GClassicVcr& vcr);
    void initVcr(const GClassicVcr& vcr);
    void doneVcr(GClassicVcr& vcr);
    bool setCapabilities(uint16 cap);
    bool playCycle();
    void playFastForward();

    int getBeamStatus(int side, int id);
    int getTorpStatus(int side, int id);
    int getTorpCount(int side);
    int getFighterCount(int side);
    int getShields(int side);
    int getDamage(int side);
    int getCrew(int side);

    int getFighterX(int side, int id);
    int getFighterStatus(int side, int id);
    int getObjectX(int side);
    long getDistance();

    VcrStatus* getStatusToken();
    void setStatus(VcrStatus& status);
};

#endif
