/**
  *  \file u/t_game_msgtemplate.cc
  *  \brief Tests for game/msgtemplate.cc
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_game.h"
#include "game/msgtemplate.h"

/** Test regular unparsed assignments. */
void
GameMsgTemplateTestSuite::testValues()
{
    // Lines for testing
    GMessageLines lines;
    lines.push_back(string_t());

    std::vector<string_t> values;
    bool result;

    GMessageTemplate tpl(MessageConfig, "foo");
    tpl.addCheckInstruction(tpl.iParse + tpl.sAny, 0, "value = $");
    tpl.addVariable("VALUE");

    // Test
    lines[0] = "value = 1";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "1");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 1);

    // Test
    lines[0] = "value = -42";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "-42");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), -42);

    // Test
    lines[0] = "value = 15%";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "15%");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 15);

    // Test
    lines[0] = "value =";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(!result);
}

/** Test regular assignments of type "X100". */
void
GameMsgTemplateTestSuite::testValuesX100()
{
    // Lines for testing
    GMessageLines lines;
    lines.push_back(string_t());

    std::vector<string_t> values;
    bool result;

    GMessageTemplate tpl(MessageConfig, "foo");
    tpl.addCheckInstruction(tpl.iParse + tpl.sAny, 0, "value = $");
    tpl.addVariable("VALUE:X100");

    // Test
    lines[0] = "value = 1";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "100");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 100);

    // Test
    lines[0] = "value = -42";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "-4200");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), -4200);

    // Test
    lines[0] = "value = 15%";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "1500");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 1500);

    // Test
    lines[0] = "value = .5";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "50");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 50);

    // Test
    lines[0] = "value = .15";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "15");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 15);

    // Test
    lines[0] = "value = .1234";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "12");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 12);

    // Test
    lines[0] = "value = 123.456";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "12345");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 12345);

    // Test
    lines[0] = "value = -123.456%";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "-12345");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), -12345);
}


/** Test regular assignments of enumerated types. */
void
GameMsgTemplateTestSuite::testValuesEnum()
{
    // Lines for testing
    GMessageLines lines;
    lines.push_back(string_t());

    std::vector<string_t> values;
    bool result;

    GMessageTemplate tpl(MessageConfig, "foo");
    tpl.addCheckInstruction(tpl.iParse + tpl.sAny, 0, "value = $");
    tpl.addVariable("VALUE:aa/bb/cc/dd");

    // Test
    lines[0] = "value = 1";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "");

    // Test
    lines[0] = "value = aa";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "0");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 0);

    // Test
    lines[0] = "value = bb";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "1");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 1);

    // Test
    lines[0] = "value = dd";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "3");
    TS_ASSERT_EQUALS(tpl.getIntValue(values[0]), 3);

    // Test
    lines[0] = "value = ddd";
    values.clear();
    result = tpl.match(lines, 1, values);
    TS_ASSERT(result);
    TS_ASSERT_EQUALS(values.size(), 1U);
    TS_ASSERT_EQUALS(values[0], "");
}
