/**
  *  \file u/t_int_compound.cc
  *  \brief Tests for compound values
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_int.h"
#include "int/interror.h"
#include "int/compound.h"
#include "int/exec.h"

/******************************** Testbed ********************************/

namespace {
    class MyCallable : public IntCallableValue {
     public:
        MyCallable()
            { ++num_instances; }
        ~MyCallable()
            { --num_instances; }

        // State inquiry
        string_t getState()
            { return state; }
        void clear()
            { state.clear(); }

        // IntCallableValue:
        virtual void call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
        virtual bool isProcedureCall();
        virtual int32_t getDimension(int32_t which);
        virtual IntContext* makeFirstContext();

        // IntValue:
        virtual string_t    toString(bool readable);
        virtual void        store(IntTagNode&, Stream&);
        virtual IntValue*   clone();

        static int num_instances;

     private:
        MyCallable(const MyCallable&);

        string_t state;
    };
    int MyCallable::num_instances;
}

void
MyCallable::call(IntExecutionContext& /*exc*/, IntDataSegment& args, bool want_result)
{
    // Fold all arguments into a string
    for (uint32_t i = 0; i < args.size(); ++i) {
        state += IntValue::toStringOf(args[i], true);
        state += ',';
    }
    state += want_result ? "y" : "n";
}

bool
MyCallable::isProcedureCall()
{
    return false;
}

int32_t
MyCallable::getDimension(int32_t which)
{
    if (which == 0) {
        return 7;
    } else {
        return 5*which;
    }
}

IntContext*
MyCallable::makeFirstContext()
{
    return 0;
}

string_t
MyCallable::toString(bool /*readable*/)
{
    return "#<MyCallable>";
}

void
MyCallable::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
MyCallable::clone()
{
    return new MyCallable();
}

/********************************** Test *********************************/

void
IntCompoundTestSuite::testClosure()
{
    // Create a test callable and make sure it works (yes, this leaks on failure)
    MyCallable* base = new MyCallable();
    TS_ASSERT_EQUALS(base->getDimension(0), 7);
    TS_ASSERT_EQUALS(base->getDimension(1), 5);
    TS_ASSERT_EQUALS(base->getDimension(7), 35);
    TS_ASSERT_EQUALS(MyCallable::num_instances, 1);

    // Try cloning
    {
        IntValue* copy = base->clone();
        TS_ASSERT_DIFFERS(base, copy);
        TS_ASSERT_EQUALS(MyCallable::num_instances, 2);
        delete copy;
        TS_ASSERT_EQUALS(MyCallable::num_instances, 1);
    }

    // Create a closure that binds no args and make sure it works
    std::auto_ptr<IntClosure> c(new IntClosure());
    c->setNewFunction(base);
    TS_ASSERT_EQUALS(MyCallable::num_instances, 1);
    TS_ASSERT_EQUALS(c->getDimension(0), 7);
    TS_ASSERT_EQUALS(c->getDimension(1), 5);
    TS_ASSERT_EQUALS(c->getDimension(7), 35);

    // Clone the closure
    {
        std::auto_ptr<IntValue> cc(c->clone());
        TS_ASSERT_EQUALS(MyCallable::num_instances, 1);
        TS_ASSERT_DIFFERS(cc.get(), c.get());
    }

    // Test call
    IntExecutionContext exc("dummy");
    {
        IntDataSegment dseg;
        dseg.pushNew(makeIntValue(1));
        dseg.pushNew(makeIntValue(9));
        dseg.pushNew(makeIntValue(5));
        c->call(exc, dseg, true);
        TS_ASSERT_EQUALS(base->getState(), "1,9,5,y");
        base->clear();
    }

    // Bind some args
    c->addNewArgument(makeIntValue(3));
    c->addNewArgument(makeStringValue("zz"));
    TS_ASSERT_EQUALS(c->getDimension(0), 5);
    TS_ASSERT_EQUALS(c->getDimension(1), 15);
    TS_ASSERT_EQUALS(c->getDimension(5), 35);

    {
        IntDataSegment dseg;
        dseg.pushNew(makeIntValue(1));
        dseg.pushNew(makeIntValue(9));
        dseg.pushNew(makeIntValue(5));
        c->call(exc, dseg, true);
        TS_ASSERT_EQUALS(base->getState(), "3,\"zz\",1,9,5,y");
        base->clear();
    }

    // Bind some more args
    {
        IntDataSegment a;
        a.pushNew(makeIntValue(999));
        a.pushNew(makeIntValue(42));
        a.pushNew(makeBoolValue(1));
        c->addNewArgumentsFrom(a, 2);
    }
    TS_ASSERT_EQUALS(c->getDimension(0), 3);
    TS_ASSERT_EQUALS(c->getDimension(1), 25);
    TS_ASSERT_EQUALS(c->getDimension(3), 35);

    {
        IntDataSegment dseg;
        dseg.pushNew(makeIntValue(1));
        dseg.pushNew(makeIntValue(9));
        dseg.pushNew(makeIntValue(5));
        c->call(exc, dseg, true);
        TS_ASSERT_EQUALS(base->getState(), "3,\"zz\",42,True,1,9,5,y");
        base->clear();
    }
}
