/**
  *  \file ui/event.h
  *  \brief Event Management Prototypes
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_EVENT_H
#define PCC_V2_UI_EVENT_H

#include <SDL_events.h>
#include <SDL_timer.h>
#include "cpluslib/types.h"

/** Our own event types. Note that code in event.cc assumes that these
    are valid SDL event codes. */
enum UIEventType {
    ev_Key        = SDL_KEYDOWN,
    ev_Mouse      = SDL_MOUSEMOTION,
    ev_Idle       = SDL_NOEVENT,

    ev_Tick       = SDL_USEREVENT      /* 8 event codes from SDL_USEREVENT available */
};

enum {
    mb_Left   = SDL_BUTTON(SDL_BUTTON_LEFT),
    mb_Right  = SDL_BUTTON(SDL_BUTTON_RIGHT),
    mb_Middle = SDL_BUTTON(SDL_BUTTON_MIDDLE),
    mb_Double = 0x8000U
};

enum {
    ss_Shift = KMOD_LSHIFT << 16,
    ss_Ctrl  = KMOD_LCTRL  << 16,
    ss_Alt   = KMOD_LALT   << 16,
    ss_Meta  = KMOD_LMETA  << 16,
    
    ss_AllMod = ss_Shift | ss_Ctrl | ss_Alt | ss_Meta
};

/** Virtual keys. These are posted as normal key events. */
enum {
    vk_WheelUp = SDLK_LAST,            ///< Mouse wheel "up".
    vk_WheelDown,                      ///< Mouse wheel "down".
    vk_Quit,                           ///< Close event, i.e.\ "X" in window corner.
    vk_Unicode,                        ///< Unicode key, i.e.\ non-standard character.
    vk_DoubleClickLocation             ///< Double-click a map location.
};

enum UIKeyClass {
    /** A normal key which is perceived as a "key press" by normal
        users. Includes printing keys and function keys. */
    KeyNormal,
    /** A key which is perceived as a "modifier key". Those are
        usually used as parts of a key combination (Alt, Ctrl, ...) */
    KeyModifier,
    /** A virtual key (vk_XXX). These are used internally by our
        program. */
    KeyVirtual
};

typedef uint32 UIKey;

/* these ought to be macros so that they can be used in `switch' */
/* #define MOD(mod) (((mod) & (KMOD_LCTRL|KMOD_LALT|KMOD_LMETA|KMOD_LSHIFT)) \
                   | (((mod)>>1) & (KMOD_LCTRL|KMOD_LALT|KMOD_LMETA|KMOD_LSHIFT))) */

/** Our own event structure. We support fewer events than SDL, but use
    a more regular format. */
struct UIEvent {
    struct Mouse {
        int   sdl_type;         // SDL event type
        int   x, y;
        int32 b;                // button state + double-click flag
    };
    /** Key event.
        - <tt>code</tt> contains the key code (SDLK_xxx, vk_xxx) and shift flags.
        - if it's printable, <tt>unicode</tt> contains the Unicode value.
          For keys outside ASCII, <tt>code</tt> is vk_Unicode. For keys
          that are not printable, this field is 0. */
    struct Key {
        int   sdl_type;
        int   arg;
        uint32 code;
        uint16 unicode;
    };
    struct Tick {
        int   elapsed;          // Number of ticks elapsed since last tick event
    };
    UIEventType type;
    union {
        /* ev_Key */
        Key     key;
        /* ev_Mouse */
        Mouse   mouse;
        /* ev_Tick */
        Tick    tick;
    };

    static inline uint32 simplMod(uint32 m) {
        return (m & ss_AllMod) | ((m >> 1) & ss_AllMod);
    }
};

struct UICommand {
    int32 code;
    int32 arg;
    int32 info, info2;
    void* infop;

    UICommand()
        : code(0), arg(0), info(0), info2(0), infop(0)
        { }
    UICommand(int32 code, int32 arg = 0, int32 info = 0)
        : code(code), arg(arg), info(info), info2(0), infop(0)
        { }
};

enum {
    TIMER_HZ = 50,                 ///< Timer frequency (Hertz).
    TIMER_MS = 1000 / TIMER_HZ     ///< Timer delay (milliseconds).
};

void getEvent(UIEvent& event, bool idle = false);
void getEvent(UIEvent& event, bool idle, bool infinite);
void putEvent(const UIEvent& event);
void putKey(uint32 code, int arg = 0);
void ungetEvent(const UIEvent& event);
void postMouseEvent();
bool haveQueuedEvent();
bool checkBreakEvent();

uint32 getCurrentModifierKeys();

long getTimer();
long getLostTicks();
void clearSkippedTicks();

void initEvents();
void doneEvents();

void setInhibitMouseGrab(bool value);
void setMouseMode(bool infinite);

UIKeyClass getClassOfKey(UIKey key);

#endif
