/**
  *  \file ui/widgets/icongrid.cc
  *  \brief UIIconGrid widget
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/icongrid.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "cpluslib/assert.h"
#include "gfx/clip.h"

/** Constructor. When the actual geometry parameters are not known at
    this point, you can pass dummy values, and call setGeometry later
    on.

    \param iconSize Size of individual icons
    \param width    Number of columns
    \param height   Number of rows
    \param id       Widget Id */
UIIconGrid::UIIconGrid(GfxPoint iconSize, int width, int height, int id)
    : UIWidget(id),
      iconSize(iconSize),
      width(width),
      height(height),
      x(0),
      y(0),
      top(0)
{ }

UIIconGrid::~UIIconGrid()
{ }

bool
UIIconGrid::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Mouse && isInside(GfxPoint(event.mouse.x, event.mouse.y)) && event.mouse.b != 0) {
        // Focus ourselves
        setFocusedWidget();

        // Move selected item
        int nx = (event.mouse.x - getExtent().x) / iconSize.x;
        int ny = (event.mouse.y - getExtent().y) / iconSize.y + top;
        if (nx < width && ny < height) {
            setSelectedWidget();
            setCurrentPos(nx, ny);
        }
        return true;
    }
    if (event.type == ev_Key && getFocusState() == PrimaryFocus) {
        switch (event.key.code) {
         case SDLK_RIGHT:
            setSelectedWidget();
            if (x < width-1)
                setCurrentPos(x+1, y);
            else if (y < height-1)
                setCurrentPos(0, y+1);
            return true;
         case SDLK_LEFT:
            setSelectedWidget();
            if (x > 0)
                setCurrentPos(x-1, y);
            else if (y > 0)
                setCurrentPos(width-1, y-1);
            return true;
         case SDLK_UP:
            setSelectedWidget();
            setCurrentPos(x, y-1);
            return true;
         case SDLK_DOWN:
            setSelectedWidget();
            setCurrentPos(x, y+1);
            return true;
         case SDLK_PAGEUP:
            setSelectedWidget();
            setCurrentPos(x, y - getExtent().h / iconSize.y);
            return true;
         case SDLK_PAGEDOWN:
            setSelectedWidget();
            setCurrentPos(x, y + getExtent().h / iconSize.y);
            return true;
        }
    }

    return UIWidget::handleEvent(event, second_pass);
}

void
UIIconGrid::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    const int nlines = (getExtent().h + iconSize.y - 1) / iconSize.y;
    const int ncols  = (getExtent().w + iconSize.x - 1) / iconSize.x;
    for (int iy = 0; iy < nlines; ++iy) {
        for (int ix = 0; ix < ncols; ++ix) {
            // Figure out location
            GfxRect area(getExtent().getTopLeft().movedBy(iconSize.scaledBy(ix, iy)), iconSize);
            GfxRect clip(area);
            clip.intersect(getExtent());

            // Can we draw this?
            int realY = iy + top;
            if (realY < height && ix < width) {
                // Valid
                GfxClipRectFilter filter(clip, can);
                drawIcon(filter, area, ix, realY, (ix == x && realY == y) ? getFocusState() : NoFocus);
            } else {
                // Not valid
                drawBackground(ctx, clip);
            }
        }
    }

    can.update(getExtent());
}

void
UIIconGrid::getLayoutInfo(LayoutInfo& info)
{
    int wi = iconSize.x * width;
    int he = iconSize.y * height;
    int limit = ui_screen->getSize().y * 3/4;
    if (he > limit)
        he = limit;

    info.min_size = GfxPoint(wi, 3 * iconSize.y);
    info.pref_size = GfxPoint(wi, he);
    info.flags = li_FlexibleV;
}

/** Set current position. Coordinates outside the valid range are moved back in.
    \param x Column [0,getNumColumns())
    \param y Line [0,getNumLines()) */
void
UIIconGrid::setCurrentPos(int x, int y)
{
    // Limit
    if (x < 0)
        x = 0;
    if (x >= width)
        x = width-1;
    if (y < 0)
        y = 0;
    if (y >= height)
        y = height-1;

    // Move y
    if (y != this->y) {
        int nlines = getExtent().h / iconSize.y;
        this->y = y;
        this->x = x;
        if (y < this->top)
            this->top = y;
        if (y >= this->top + nlines)
            this->top = y - nlines+1;
        drawWidget();
        sig_moved.raise();
    }

    // Move x
    if (x != this->x) {
        this->x = x;
        drawWidget();
        sig_moved.raise();
    }
}

/** Set current linear position.
    \param n Linear position. [0,getNumColumns()) for first line, [getNumColumns(),2*getNumColumns())
             for second line, and so on */
void
UIIconGrid::setCurrentPos(int n)
{
    setCurrentPos(n % width, n / width);
}

/** Set geometry parameters. May only be called during initialisation
    (construction), when the widget is not yet visible. */
void
UIIconGrid::setGeometry(GfxPoint iconSize, int width, int height)
{
    ASSERT(!hasState(st_Exposed));
    this->iconSize = iconSize;
    this->width    = width;
    this->height   = height;
}
