/**
  *  \file util/xmlread.h
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_XMLREAD_H
#define PCC_V2_UTIL_XMLREAD_H

#include <cstddef>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "util/recode.h"

class Stream;

/** Simple XML Reader. This is a lexical analyzer for XML. It dissects
    an XML document into tokens. It is intended for parsing XML files
    of known formats. Input can have UTF-8, UTF-16LE or UTF16-BE
    encoding, or one of the supported codepages. Output is always UTF-8.

    \todo This does not support DTDs. Externally-referenced DTDs are
    not read, embedded DTDs are not parsed (and may actually be
    misparsed if present); entities defined in a DTD are thus not
    expanded. */
class XmlReader {
 public:
    enum Token {                // extra data        tag         name         value
        Eof,                    // end of file;
        TagStart,               // starting tag;     tag name
        TagAttribute,           // attribute;        tag name    attrib name  attrib value
        TagEnd,                 // ending tag;       tag name
        PIStart,                // processing insn;  insn name
        PIAttribute,            // PI attribute;     insn name   attrib name  attrib value
        Comment,                // comment
        Text,                   // text;                                      text
        Null,                   // internal
        Error                   // syntax error
    };
    enum WhitespaceMode {
        TrimWS,                 // Remove text nodes containing just whitespace
        SingleWS,               // Trim text nodes containing just whitespace to a single space
        AllWS                   // Keep text nodes containing just whitespace entirely
    };

    XmlReader(Stream& s);
    virtual ~XmlReader();

    Token read();
    string_t getTag() const;
    string_t getName() const;
    string_t getValue() const;

    WhitespaceMode getWhitespaceMode() const;
    void setWhitespaceMode(WhitespaceMode mode);

    virtual string_t expandEntity(string_t e);

 private:
    enum Encoding {
        Utf8,
        Codepage,
        Utf16LE,
        Utf16BE
    };
    enum State {
        Main,                   // regular state
        InTag,                  // after reading an opening tag
        InPI                    // after reading an opening PI
    };
    enum {
        WhitespaceChar = 1,
        LTChar         = 2,
        GTChar         = 4,
        EQChar         = 8,
        BangChar       = 16,
        QuesChar       = 32,
        SlashChar      = 64,
        MinusChar      = 128,
        LBracketChar   = 256,
        AnyChar        = 512,

        ProcessQuotes  = 16384
    };

    Stream&      s;
    std::size_t  buf_ptr;
    std::size_t  buf_fill;
    bool         eof;
    char         buffer[1024];
    Encoding     encoding;
    CharacterSet codepage;
    State        state;
    int32_t      current_char;
    WhitespaceMode ws_mode;

    string_t     tag;           // name of current tag
    string_t     name;          // name of current attribute
    string_t     value;         // value or text

    Token    readNextToken();
    void     readNextChar();
    int      readByte();
    void     refillBuffer();
    void     skip(std::size_t n);
    uint8_t  peek(std::size_t n);
    void     detectEncoding();
    int32_t  readCharacterSequence(int cl, string_t* accum);
    void     readUnparsedCharacterData(string_t& out);
    string_t expandEntities(string_t s);
};

#endif
